:- import append/3, member/2 from basics.
:- import subsumes/2 from subsumes.

:- dynamic failed_pred/1.

cleanup_clauses(CList, Rest) :-
	cleanup_flag(Cleanup),
	( member(subsumed, Cleanup) 
		-> delete_subsumed_clauses(CList, New)
		;  New = CList ),
	( member(failed, Cleanup) 
	        -> delete_failed_clauses(New, Rest)
		;  Rest = New
        ).

delete_subsumed_clauses(CList, New) :-
	( get_subsumed_clause(CList, N) 
		-> retractall(clauseI(N, _, _)),
		   remove(N, CList, CList1), 
	           delete_subsumed_clauses(CList1, New)
		;  New = CList
        ).


get_subsumed_clause(CList, N) :- get_subsumed_clause(CList, CList, N).

get_subsumed_clause([N|Rest], CList, Subs_clause) :-
	( subsumed_clause_chk(N, CList) 
		-> Subs_clause = N
		;  get_subsumed_clause(Rest, CList, Subs_clause)
        ).


% check whether clause N is subsumed by any other clause in CList
% calling mode : (+, +)

subsumed_clause_chk(N, CList) :-
	member(M, CList), M \= N,
	clauseI(M, (H :- B), _), 
	clauseI(N, (Hd :- Bd), _),
	copy_term((H :- B), (H1 :- B1)),
	copy_term((Hd :- Bd), (Hd1 :- Bd1)),
	partition(Bd1, X, _),
	subsumes((H1 :- B1), (Hd1 :- X)).

% some old code : not deleting it : will check whether it is "dead"
% clause_subscheck(C, D) 
%       True if clause C subsumes clause D

clause_subscheck((H :- B), (Hd :- Bd)) :-
	copy_term((H :- B), (H1 :- B1)),
	copy_term((Hd :- Bd), (Hd1 :- Bd1)),
	partition(Bd1, X, _),
	subsumes((H1 :- B1), (Hd1 :- X)).
	
%----------------------------------------------------------------

% deleting the failed clauses of a predicate : assume CList is
% the list of all the clauses of some program predicate.

delete_failed_clauses(CList, Rest) :-
	( delete_pred(CList) -> Rest = []
		             ;  remove_failed_clauses(CList, Rest)
        ).

remove_failed_clauses([], []).
remove_failed_clauses([C|Rest], List) :-
	get_clause(C, (_ :- Bd), _),
	( failed_body(Bd) -> retract_clauses([C]), 
			     Non_failed = []
			  ;  Non_failed = [C]
        ), remove_failed_clauses(Rest, List1),
	append(Non_failed, List1, List).
	

delete_pred(CList) :-
	CList = [Id|_], clauseI(Id, (Hd :- _), _), 
	functor(Hd, P, N),
	failed_clauses(P/N, CList),
	retract_clauses(CList),
	assert(failed_pred(P/N)),
	propagate_failures([P/N]).

failed_clauses(_/_, []).
failed_clauses(P/N, [Id|Rest]) :-
	clauseI(Id, (_ :- Body), _),
	failed_body(Body, P/N),
	failed_clauses(P/N, Rest).

failed_body(Body) :- 
	member(Atom, Body), 
	functor(Atom, Q, M),
        ( Q/M = fail/0 ; Q/M = false/0 ; failed_pred(Q/M)
        ).	

failed_body(Body, P/N) :-
	member(Atom, Body), 
	functor(Atom, Q, M),
        ( Q/M = P/N ; Q/M = fail/0 ; 
          Q/M = false/0 ; failed_pred(Q/M)
        ).

exists_clause(P/N) :- 
	clauseI(_, (Hd :- _), _), functor(Hd, P, N).

%------------------------------------------------------------------

propagate_failures([]).

propagate_failures([P/N | Tail]) :-
	non_empty_preds(PredList1),
	remove_occurrences(P/N),
	non_empty_preds(PredList2),
	set_diff(PredList1, PredList2, Rest),
	append(Tail, Rest, Rest1),
	propagate_failures(Rest1).

remove_occurrences(P/N) :-
	clauseI(X, (_ :- Bd), _),
	member(A, Bd), functor(A,P,N),
	retractall(clauseI(X, _, _)), fail.

remove_occurrences(_/_).