
:- import variant/2 from subsumes.
:- import append/3, member/2, length/2 from basics.

/*
goal_replace(Conj1/HV1, Conj2/HV2, Clause1, NewId) :-
	var_seteq(HV1, HV2),
	create_defn(Conj1, HV1, Pred1),
	create_defn(Conj2, HV2, Pred2),
	prove(Pred1, Pred2, D1, D2),!,
	update_proof_bounds(Conj1, Conj2, (D1, D2), (Delta1, Delta2)),
	Clause1 = clauseI(CId, Hd_Bd, (An1, An2)),
	gr_annot_chk(An1, Delta1, Hd_Bd),
	replace(Hd_Bd, Conj1, Conj2, Hd_Bdnew),
	plus_annot(An1, Delta1, AD1),
	plus_annot(An2, Delta2, AD2),
	retractall(clauseI(CId, _, _)), 
	gen_new_clauseNum(NewId),
	assert(clauseI(NewId, Hd_Bdnew, (AD1, AD2))).
*/
%---------------------------------------------------------------------

create_defn(Goal, HVars, Pred) :-
	exists_defn(Goal, HVars, Pred), !.

create_defn(Goal, HVars, Pred) :-
	( open_atom(Goal) -> atom_pred(Goal, P, N), Pred = P/N
			  ;  length(HVars, N), get_new_pred(P), 
                             Pred = P/N, Head =.. [P|HVars], 
                             define(P, N, [(Head :- Goal)])
        ).

exists_defn(Goal, HVars, Pred) :-
	clause0(C, (H :- B), _),
	functor(H, P, N), length(L, N), Hd =.. [P|L],
	findall(Cl, clause0(Cl, (Hd :- _), _), [C]),
	Head =.. [P|HVars], copy_term((H :- B), (Hnew :- Bnew)),
	exec_builtins(Goal, Goal1), exec_builtins(Bnew, B1),
        my_variant((Head :- Goal1), (Hnew :- B1)), Pred = P/N.

atom_pred(G, P, N) :-
	vanillaConvert(G, G1), functor(G1, P, N).

open_atom(G) :- vanillaConvert(G, G1), G1 =.. [_|L], var_list(L).

var_list([]).
var_list([V|Rest]) :- var(V), var_list(Rest).
	
define(P, Arity, CList) :-
	CList = [_|_], 
	syntax_check(P, Arity, CList),
        define_clauses(CList).

syntax_check(_, _, []).
syntax_check(P, A, [C|Rest]) :-
	C = (H :- _), functor(H, P, A), syntax_check(P, A, Rest).


define_clauses([]).
define_clauses([C|Rest]) :-
	gen_new_clauseNum(N), C = (H :- B), 
	assert_def(N, (H :- B), (1,1)),
	define_clauses(Rest).

%---------------------------------------------------------------------

update_proof_bounds(A1, A2, (D1, D2), (D1, D2)) :-
	open_atom(A1), open_atom(A2), !.

update_proof_bounds(A1, A2, (D1, D2), (D1, D2)) :-
	\+ open_atom(A1), \+ open_atom(A2), !.

update_proof_bounds(A1, A2, (D1, D2), (Delta1, Delta2)) :-
	\+ open_atom(A1), open_atom(A2), !,
	minus_annot(D1, 1, Delta1), 
	minus_annot(D2, 1, Delta2). 

update_proof_bounds(A1, A2, (D1, D2), (Delta1, Delta2)) :-
	open_atom(A1), \+ open_atom(A2),
	plus_annot(D1, 1, Delta1), 
	plus_annot(D2, 1, Delta2). 

%-----------------------------------------------------------------
% the third condition for total correctness of Goal Replacement 
% given in the PPDP '99 paper. This condition allows us to prove 
% the total correctness of Goal Replacement by induction.

gr_annot_chk(Gamma, Delta, (_ :- Bd)) :-
	length(Bd, N), K is N - 1,       % replacing a single atom
	plus_annot(Gamma, Delta, GD),
	plus_annot(GD, K, GDK),
	greater_annot(GDK, 0).

%-----------------------------------------------------------------------

replace((Hd :- Bd), [], [], (Hd :- Bd)).

replace((Hd :- Bd), [A|Rest], [A1|Rest1], (Hd :- Bd1)) :-
	search_and_replace(Bd, A, A1, Bd_tmp),
	replace((Hd :- Bd_tmp), Rest, Rest1, (Hd :- Bd1)).

search_and_replace([A1|Rest1], Atom1, Atom2, [A2|Rest2]) :-
	( variant(A1, Atom1) 
                -> A1 = Atom1, A2 = Atom2, Rest2 = Rest1
                ;  A2 = A1,
                   search_and_replace(Rest1, Atom1, Atom2, Rest2)
        ).
	

%---------------------------------------------------------------
	
cond_axiom(Pred1, Pred2) :-
	( cleanup_and_axiom_chk(Pred1, Pred2)
		-> true
		;  find_clauses(Pred1, CL1),
		   find_clauses(Pred2, CL2),
		   gen_prove_conds(CL1, CL2),
	           cleanup_and_axiom_chk(Pred1, Pred2)
        ).


%  generate and prove conditions ; 
%  it is possible to exercise considerable choice in
%  finding the conditions to be proved; because a syntactic equiv.
%  check is performed before cond_axiom succeeds.

gen_prove_conds([], _).

gen_prove_conds([C|Rest1], Cq) :-
	findall(D, match_clause(C,Cq,D), List),
	(null_condition(C, List) 
		-> true
	        ;  ( List = [D] 
		        -> retrieve_pred(C, New1),
		           retrieve_pred(D, New2),
		           prove(New1, New2)
                        ;  ( List = []
		               -> retrieve_pred(C, NewPred/N),
		                  prove(NewPred/N, false/N)
             		       ;  writeln('matching with multiple clauses'),
		                  writeln('Disjunctive defns not handled'), fail
                           )
                   )
        ), remove_all(List, Cq, Cq_rest),
	gen_prove_conds(Rest1, Cq_rest).
             

%------------------------------------------------------------

match_clause(C, Cq, D) :-
	get_clause(C, (H1 :- B1), _),
	member(D, Cq), 
	get_clause(D, (H2 :- B2), _),
	exec_builtins(B1, _), 
	exec_builtins(B2, _),
	head_check(H1, H2).


retrieve_pred(C, New) :-
	get_clause(C, (H1 :- B1), _),
	exec_builtins(B1, Bd1), 
	get_vars(H1, HVars1),		
	create_defn(Bd1, HVars1, New),
	defn_fold(C, New).


head_check(H1, H2) :-
	H1 =.. [_|L1], H2 =.. [_|L2], my_variant(L2, L1).


defn_fold(C, Pred) :-
	Pred = P/N, get_clause(C, (_ :- Bd), _),
	Bd = [Atom], length(Args, N), Atom =.. [P|Args], !.

defn_fold(C, Pred) :- fold([C], Pred, _).


null_condition(C, []) :-
	get_clause(C, (H1 :- B1), _),
	exec_builtins(B1, Bd1),	
	( my_variant(Bd1, false) ; my_variant(Bd1, fail)
	  ; ( get_vars(H1, HV1),
	      create_defn(Bd1, HV1, New/N),
	      proof_stack(New/N, false/N),
              defn_fold(C, New/N)
            )
        ).

null_condition(C, [D]) :-
	get_clause(C, (H1 :- B1), _),
	get_clause(D, (H2 :- B2), _),
	exec_builtins(B1, Bd1),
	exec_builtins(B2, Bd2),
	( ( H1 =.. [_|L1], H2 =.. [_|L2],
	    my_variant(Bd1/L1, Bd2/L2)
          ) ;
	  ( get_vars(H1, HV1),
	    get_vars(H2, HV2),
	    create_defn(Bd1, HV1, New1),
	    create_defn(Bd2, HV2, New2), 
	    proof_stack(New1, New2),
	    defn_fold(C,New1), defn_fold(D,New2)
          )
        ).
