:- import member/2, length/2 from basics.

same(Pred, X, Y) :-
	find_clauses(Pred, CL),
	member(X, CL), member(Y, CL),
	get_clause(X, (Hx :- _), _),
	get_clause(Y, (Hy :- _), _), 
	\+(X == Y), my_variant(Hx, Hy).


nexists(Pred1, Pred2, X) :-
	find_clauses(Pred1, CL),
	find_clauses(Pred2, DL),
	member(X, CL), get_clause(X, (Hx :- _), _),
	not( (member(Y, DL), 
	     get_clause(Y, (Hy :- _), _),
	     Hx =.. [_|Lx], Hy =.. [_|Ly], 
	     my_variant(Lx, Ly))
           ).


cleanup(Pred, N) :-
	find_clauses(Pred, CL), 
	cleanup_clauses(CL, DL),
	length(DL, N).

dump_pred(File, Pred1, Pred2) :-
	Pred1 = P/N, Pred2 = Q/M,
	tell(File), write_all(P, N),
	nl, write_star(32), nl,
	write_all(Q, M), told.


pairup(Pred1, Pred2, CL1, DL1) :-
	find_clauses(Pred1, CL),
	find_clauses(Pred2, DL),
	tell('ill_dump1'),
	paircl(CL, DL, CL1, DL1), told.


paircl([], DL, [], DL).

paircl([X|T], DL, CL1, DL1) :-
	get_clause(X, (Hx :- _), _),
	( get_match_head(Hx, DL, Y) 
		-> remove(Y, DL, DL_rest),
		   write_clause(X), nl, 
		   write_clause(Y), nl, write_star(32), nl,
		   paircl(T, DL_rest, CL1, DL1)
		;  paircl(T, DL, CL2, DL2),
		   CL1 = [X|CL2], DL1 = DL2
        ). 

get_match_head(Hx, DL, Y) :-
	member(Y, DL), 
	get_clause(Y, (Hy :- _), _),
	Hx =.. [_|Lx], Hy =.. [_|Ly], 
	my_variant(Lx, Ly).

get_defn(P/N, CList) :-
	length(L, N), Hd =.. [P|L],
	findall(C, clause0(C, (Hd :- _), _), CList),
	write_clauses(CList).


print_tree(Pred1, Pred2) :-
	write('BEG of DAG : '), print_node(Pred1, Pred2),
	findall((X, Y), oblig_arc(Pred1, Pred2, X, Y), Children),
	print_trees(Children), write('END of DAG : '),
	print_node(Pred1, Pred2).

print_trees([]).
print_trees([H|T]) :- 
	H = (X, Y), print_tree(X, Y), print_trees(T).

print_node(X, Y) :-
	write(X), write(' , '), write(Y), write(' , '),
	( proved(X, Y) -> writeln(success) 
		       ;  writeln(failure)
        ).


add_all_deduce(T) :- 
	findall((T1, T2), deduce_time(T1, T2), TList),
	add_all(0, TList, T).

add_all(N, [], N).

add_all(N, [(H1,H2)|T], M) :- N1 is N + H2 - H1, add_all(N1, T, M).

write_oblig :-
	tell('meta_oblig'),
	write_oblig_all, told.

write_oblig_all :- 
	oblig_arc(X, Y, Z, W),
	write(arc((X, Y), (Z, W))),
	writeln('.'), fail.

write_oblig_all.

oblig_depth(N, K) :- 
	findall(C, arc(N, C), CL),
	oblig_maxdepth(CL, K1), 
	K is K1+1.

oblig_maxdepth([], 0).

oblig_maxdepth([N|Rest], K) :-
	oblig_depth(N, K1),
	oblig_maxdepth(Rest, K2),
	( K1 > K2 -> K = K1
		  ;  K = K2
        ).


